#From the paper:
#Evolutionary Clustering and Community Detection Algorithms for Social Media Health Surveillance
#Heba Elgazzar, Kyle Spurlock, Tanner Bogart
#2020

import numpy as np
#Harris, C.R., Millman, K.J., van der Walt, S.J. et al. Array programming with NumPy. Nature 585, 357–362 (2020). DOI: 0.1038/s41586-020-2649-2. (Publisher link).
import matplotlib.pyplot as plt
#Hunter, J. D. (2007). Matplotlib: a 2D graphics environment. Computing in Science & Engineering, 9(3), 90-95. doi:10.1109/mcse.2007.55.
from sklearn.metrics import pairwise_distances
#Pedregosa, F., Varoquaux, G., Gramfort, A., Michel, V., Thirion, B., Grisel, O., . . . Duchesnay, E. (2011). Scikit-learn: machine learning in python (M. Braun, Ed.). The Journal of Machine Learning Research, 12, 2825-2830. Retrieved November 16, 2020, from jmlr.org.
import community as community_louvain
#Aynaud, T. (2020). python-louvain x.y: Louvain algorithm for community detection. GitHub. https://github.com/taynaud/python-louvain.
import networkx as nx
#Hagberg, A., Swart, P., & S Chult, D. (2008, August). Exploring network structure, dynamics, and function using networkx. Proceedings of the 7th Python in Science Conference, 11-16. Retrieved November 16, 2020, from www.osti.gov.

class EvoLouvain:
    def __init__(self): #constructor
        self.current_time = 0 #the current generation step of the class
        self.previous_gen = [] #holds the previous generations distance matrix
        self.modularities = [] #Modularities per timesteps
        self.times = []
        
    def showPlot(self, current_gen, partition, modularity, alpha): #Displays the plots of the clustered generation
        plt.scatter(current_gen[:,0], current_gen[:,1], c = list(partition.values()), cmap = 'tab20') #(x, y, cluster_ labels, colour_map for clusters)
        title = r'Louvain Generation-{g} Alpha-{a} Modularity-{m}'.format(g = self.current_time, a = alpha, m = "%.5f" % modularity)
        plt.title(title)
        plt.savefig('{t}.png'.format(t = title))
        plt.show() #Show the graph
    
    def applySmoothing(mat1, mat2, alpha): #Adding two matrices
        mat1, mat2 = (1-alpha)*mat1, alpha*mat2
        ysize, xsize = mat2.shape #Get inverse of shape of second matrix
        mat1[0:ysize, 0:xsize] += mat2 #Add elements of mat2 to mat1 on index (0,0)
        return mat1
    
    def callLouvain(self, X, t_intervals, alpha, show_mod = False): #member function that performs the evolutionary clustering
        seen_times = [] #these are the time steps we have currently moved through
        for time in t_intervals: #For each time step in our unique times
            self.current_time = time #Set the class time to this time
            seen_times.append(time) #Add the current_time to our already viewed times
            
            current_gen = X.loc[X['Time'].isin(seen_times)] #Check which values in our data match the current time step
            current_gen = current_gen.iloc[:,[0,1]].values #Convert the dataframe with the time column into a 2-D array with the coordinate points
            
            dist_cg = pairwise_distances(current_gen) #Calculate the distance matrix of our coordinate points using pairwise distance
            dist_cg = EvoLouvain.convert_matrix(dist_cg) #Remove distant links
            
            if time == 0: #If we are at the first step, there is no previous generation to consider
                dist_cg = dist_cg #Otherwise take distance matrix as normal
            else: #If we are past the first time step, there is a previous generation we must consider
                #dist_cg = EvoLouvain.addAtPos(((1-alpha)*dist_cg), ((alpha)*self.previous_gen))
                dist_cg = EvoLouvain.applySmoothing(dist_cg, self.previous_gen, alpha)
            
            G = nx.from_numpy_matrix(dist_cg) #Get NetworkX graph
            partition = community_louvain.best_partition(G)#Find the best partition at time step
            
            modularity = 0
            try:
                modularity = community_louvain.modularity(partition, G) #Retrieve modularities
            except(ValueError):
                print("Can't compute modularity for this partition.")
            
            if (show_mod):
                print('Generation {g} modularity: {m}'.format(g = self.current_time, m = modularity))

            self.modularities.append(modularity) #Store these for possible visualization later
            self.times = seen_times
            
            self.previous_gen = dist_cg #Set the current generations distance matrix to the previous gen since we have passed it

            if time in [100, int(np.median(t_intervals)), t_intervals[-1]]: #If the time has either started, is in the middle, or at the end:
                EvoLouvain.showPlot(self, current_gen, partition, modularity, alpha) #Take our current gen coordinates and the predicted labels and plot them
    
    def convert_matrix(dist): #Introducing sparsity into distance matrix, loose similarity
        dist[(dist < 10) & (dist != 0)] = 3
        dist[(dist > 10) & (dist < 20)] = 2
        dist[(dist > 20) & (dist < 30)] = 1
        dist[dist > 30] = 0
        return dist


#From the paper:
#Evolutionary Clustering and Community Detection Algorithms for Social Media Health Surveillance
#Heba Elgazzar, Kyle Spurlock, Tanner Bogart
#2020

#This collection app includes preprocessing that masks the identity of the collected users.
#Twitter API credentials may also not be disclosed based on Twitter Developer terms & conditions.

import numpy as np
#Harris, C.R., Millman, K.J., van der Walt, S.J. et al. Array programming with NumPy. Nature 585, 357–362 (2020). DOI: 0.1038/s41586-020-2649-2. (Publisher link).
import pandas as pd
#McKinney, W. (2010, June). Data structures for statistical computing in Python. Proceedings of the 9th Python in Science Conference, 445, 51-56. doi:10.25080/majora-92bf1922-00a.
import matplotlib.pyplot as plt
#Hunter, J. D. (2007). Matplotlib: a 2D graphics environment. Computing in Science & Engineering, 9(3), 90-95. doi:10.1109/mcse.2007.55.
from sklearn.cluster import DBSCAN
from sklearn.metrics import pairwise_distances
from sklearn.preprocessing import LabelEncoder
#Pedregosa, F., Varoquaux, G., Gramfort, A., Michel, V., Thirion, B., Grisel, O., . . . Duchesnay, E. (2011). Scikit-learn: machine learning in python (M. Braun, Ed.). The Journal of Machine Learning Research, 12, 2825-2830. Retrieved November 16, 2020, from jmlr.org.
import community as community_louvain
#Aynaud, T. (2020). python-louvain x.y: Louvain algorithm for community detection. GitHub. https://github.com/taynaud/python-louvain.
import networkx as nx
#Hagberg, A., Swart, P., & S Chult, D. (2008, August). Exploring network structure, dynamics, and function using networkx. Proceedings of the 7th Python in Science Conference, 11-16. Retrieved November 16, 2020, from www.osti.gov.
import tweepy
#Roesslein, J. (2009). Tweepy documentation. Retrieved November 16, 2020, from https://docs.tweepy.org/en/v3.5.0/
import pkg_resources

#Requires Twitter API keys
#Retrieve from here https://developer.twitter.com
consumer_key = ''
consumer_secret_key = ''
access_token = ''
access_token_secret = ''
###############################################################################

auth = tweepy.OAuthHandler(consumer_key, consumer_secret_key)
auth.set_access_token(access_token, access_token_secret)
api = tweepy.API(auth, wait_on_rate_limit = True, wait_on_rate_limit_notify=True)

user_list = [] #Lists to hold various user characteristics, 
city_list = [] #later to be included within csv file.
state_list = []
lat_list = []
long_list = []
time_list = []

seen_users = [] #Ensuring the same user is not being picked up multiple times.

try:
    data = pkg_resources.resource_stream(__name__, 'data/uscities.csv')
    uscities = pd.read_csv(data).iloc[:,[0,3,6,7]]
except FileNotFoundError as e:
    print(e)
    print("Could not include uscities.csv for TweepyStream module.")

#(2020). United States Cities Database. SimpleMaps. https://simplemaps.com/data/us-cities.

###############################################################################

def stream_tweets(search_term, n_iter = 1000):
    counter = 0 # counter to keep track of each iteration
    for tweet in tweepy.Cursor(api.search, q='\"{}\" -filter:retweets'.format(search_term), count=10, lang='en').items():
            if counter == n_iter:
                break
            
            else:
                if (tweet.user.location and (tweet.user.id_str not in seen_users)
                    and 40 < tweet.user.followers_count < 10000):
                    if (location_verify(tweet.user.location, uscities)): #If user's location is enabled and verified
                        print('Found')
                        loc = tweet.user.location
                        loc = location_change(loc, uscities) #Perform conversion of user location using uscities database.
                        seen_users.append(tweet.user.id_str)
                        user_list.append(tweet.user.id_str)
                        city_list.append(loc[0])
                        state_list.append(loc[1])
                        lat_list.append(loc[2])
                        long_list.append(loc[3])
                        time_list.append(tweet.created_at)
                        
                        counter+=1
    return 1

def location_verify(location, uscities): #Verifies that the location for the user actually exists
    for index, city in uscities['city'].iteritems():
        if location.find(city) != -1:
            return True
    return False
            

def location_change(location, uscities): #Expands verified location to include state and coordinate elements
    for index, city in uscities['city'].iteritems():
        if location.find(city) != -1:
            loc_line = [str(uscities['city'][index]), str(uscities['state_name'][index]), str(uscities['lat'][index]), str(uscities['lng'][index])]
            return loc_line
        
def full_preprocess(df):
    lbe = LabelEncoder()

    df['User'] = lbe.fit_transform(df['User'])
    
    T = df['Timestamp'].to_list() #Take the timestamp column again and turn it into a list
    t = [] #List to hold time converted to seconds
    
    for i in T: #For item in the timestamps
        i = str(i)
        year, daytime = i.split(' ') #Split the year/month from the time of day
        hour, minute, sec = daytime.split(':') #Split the time of day into hours, minutes, and seconds
        time = int(minute)*60+int(sec) #Change the time to seconds, not using hours because it won't matter
        t.append(time) #Add the converted time in seconds to the t list
        
    lbe = LabelEncoder() #LabelEncoder class object
    t = lbe.fit_transform(t) #Encode the time in second values 
    
    df['Timestamp'] = t #Replace the timestamps of our data with the new encoded values
    
    df = (df.sort_values(by=['Timestamp'], ascending = True)).values #Sort these data points by the smallest time to the longest time
    
    df = pd.DataFrame(df, columns=['ID','City','State','Lat','Long','Time'])
    
    df.to_csv('twitter_dataset_new.csv', index=False)


#From the paper:
#Evolutionary Clustering and Community Detection Algorithms for Social Media Health Surveillance
#Heba Elgazzar, Kyle Spurlock, Tanner Bogart
#2020

import numpy as np
#Harris, C.R., Millman, K.J., van der Walt, S.J. et al. Array programming with NumPy. Nature 585, 357–362 (2020). DOI: 0.1038/s41586-020-2649-2. (Publisher link).
import matplotlib.pyplot as plt
#Hunter, J. D. (2007). Matplotlib: a 2D graphics environment. Computing in Science & Engineering, 9(3), 90-95. doi:10.1109/mcse.2007.55.
from sklearn.cluster import DBSCAN
from sklearn.metrics import pairwise_distances
#Pedregosa, F., Varoquaux, G., Gramfort, A., Michel, V., Thirion, B., Grisel, O., . . . Duchesnay, E. (2011). Scikit-learn: machine learning in python (M. Braun, Ed.). The Journal of Machine Learning Research, 12, 2825-2830. Retrieved November 16, 2020, from jmlr.org.

class EvoDBSCAN:
    def __init__(self):
        self.current_time = 0 #The current generation step of the class
        self.previous_gen = [] #Holds the previous generations distance matrix
        self.clusters = [] #Number of clusters
        self.noise = [] #Number of noise points
    
    def showPlot(self, current_gen, labels, noise, alpha): #Displays the plots of the clustered generation
        plt.scatter(current_gen[:, 0], current_gen[:, 1], c = labels, cmap = 'tab20') #(x, y, cluster_labels, colour_map for clusters)
        plt.scatter(current_gen[labels == -1, 0], current_gen[labels == -1 , 1], c = 'black') #Plot noise
        title = 'EvoDBSCANv1 Generation-{t} Alpha-{a} Noise-{n}'.format(t = self.current_time, a = alpha, n = noise)
        plt.title(title) #Title of the graph, displays the current generation step
        plt.savefig('{t}.png'.format(t = title))
        plt.show() #Show the graph
    
    def callDBSCAN(self, X, t_intervals, alpha, beta, show_eps = False): #Member function that performs the evolutionary clustering
        seen_times = [] #These are the time steps we have currently moved through
        
        for time in t_intervals: #For each time step in our unique times
            eps = 0 #Initialize epsilon (radius measure)
            
            self.current_time = time #Set the class time to this time
            seen_times.append(time) #Add the current_time to our already viewed times
            
            current_gen = X.loc[X['Time'].isin(seen_times)] #Check which values in our data match the current time step
            current_gen = current_gen.iloc[:,[0,1]].values #Convert the dataframe with the time column into a 2-D array with the coordinate points 
            
            dist_cg = pairwise_distances(current_gen) #Calculate the distance matrix of our coordinate points using pairwise distance
            
            if time == 0: #If we are at the first step, there is no previous generation to consider for our epsilon value
                if len(current_gen < 2): #If there is only one sample at first gen, median will be 0 since distance will be 0
                    eps = 1e-5 #Small value to avoid skewing radius
                else:
                    eps = ((np.median(np.unique(dist_cg))/beta)) #Find the median distance in the list of unique distances and divide it by 10
            else: #If we are past the first time step, there is a previous generation we must consider
                eps = ((1-alpha)*np.median(np.unique(dist_cg)))/beta + (alpha*np.median(np.unique(self.previous_gen)))/beta #Calculate the epsilon using both the current median unique distance, as well as the previous generation's distance
            
            if show_eps:
                print('Generation {g} epsilon: {e}'.format(g = self.current_time, e = eps))
            
            dbscan = DBSCAN(eps, min_samples = 2) #Create the dbscan object
            labels = dbscan.fit_predict(current_gen) #Find the cluster labels of our current generation
            noise = list(labels).count(-1)
            
            self.clusters.append(len(set(labels)) - (1 if -1 in labels else 0))
            self.noise.append(noise)
            
            self.previous_gen = dist_cg #Set the current generations distance matrix to the previous gen since we have passed it
            
            if time in [100, int(np.median(t_intervals)), t_intervals[-1]]: #If the time has either started, is in the middle, or at the end:
                EvoDBSCAN.showPlot(self, current_gen, labels, noise, alpha) #Take our current gen coordinates and the predicted labels and plot them

    def callSTATIC(self, X, beta): #Normal DBSCAN where epsilon is static and points are taken at once
        X = X.iloc[:,[0,1]].values
        dist_cg = pairwise_distances(X)
        eps = np.median(np.unique(dist_cg))/beta
        
        dbscan = DBSCAN(eps, min_samples = 2)
        labels = dbscan.fit_predict(X) #Find the cluster labels of our current generation
        self.clusters = len(set(labels)) - (1 if -1 in labels else 0)
        self.noise = list(labels).count(-1)
        
        #Plotting
        plt.scatter(X[:,0], X[:,1], c = labels, cmap = 'tab20') #(x, y, cluster_labels, colour_map for clusters)
        plt.scatter(X[labels == -1, 0], X[labels == -1 , 1], c = 'black') #Plot noise
        title = 'Full Static EvoDBSCANv1 Noise-{n}'.format(n = self.noise)
        plt.title(title)
        plt.savefig('{t}.png'.format(t = title))
        plt.show()


from setuptools import setup, find_packages

VERSION = '1.1.1' 
DESCRIPTION = 'Evolutionary DBSCAN and Louvain Method with a Twitter data stream'
LONG_DESCRIPTION = open("README.md").read()
INSTALL_REQUIRES = ['numpy>=1.20.3',
                    'pandas>=1.2.5',
                    'scikit-learn>=0.24.2',
                    'matplotlib>=3.4.2',
                    'python-louvain>=0.13',
                    'tweepy==3.8.0',
                    'networkx>=2.6.1']

setup(
        name="evoclusterstream",
        version=VERSION,
        url="https://github.com/kspurlock/EvoClusterStream",
        author="Kyle Spurlock, Heba Elgazzar, Tanner Bogart",
        author_email="<kdspurlock@moreheadstate.edu>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=INSTALL_REQUIRES,
        license='MIT',
        keywords=["Unsupervised machine learning", "evolutionary clustering",
                  "social networks", "Python", "DBSCAN", "Louvain method"],
        classifiers= [
        'Programming Language :: Python :: 3',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
        ],
        include_package_data=True,
        package_data={'':['data/*.csv']},
)